"""Unit test related to the generator building and feeding
"""

import numpy as np
import pytest

from deeposlandia import AVAILABLE_MODELS, generator, utils


def test_feature_detection_labelling_concise():
    """Test `feature_detection_labelling` function in `generator` module by
    considering a concise labelling, *i.e.* all labels are represented into the
    array:
    * as a preliminary verification, check if passing string labels raises an
    AttributeError exception
    * test if output shape is first input shape (batch size) + an additional
    dimension given by the `label_ids` length
    * test if both representation provides the same information (native array
    on the first hand and its one-hot version on the second hand)
    """
    array = np.array(
        [
            [
                [[10, 10, 200], [10, 10, 200], [10, 10, 200]],
                [[200, 200, 200], [200, 200, 200], [10, 10, 200]],
                [[200, 200, 200], [200, 200, 200], [200, 200, 200]],
            ],
            [
                [[10, 200, 10], [10, 200, 10], [10, 10, 200]],
                [[200, 10, 10], [10, 200, 10], [10, 10, 200]],
                [[10, 200, 10], [200, 10, 10], [10, 10, 200]],
            ],
        ]
    )
    labels = np.unique(array.reshape(-1, 3), axis=0).tolist()
    wrong_config = [
        {"id": "0", "color": [10, 10, 200], "is_evaluate": True},
        {"id": "1", "color": [200, 10, 10], "is_evaluate": True},
        {"id": "2", "color": [10, 200, 10], "is_evaluate": True},
        {"id": "3", "color": [200, 200, 200], "is_evaluate": True},
    ]
    with pytest.raises(ValueError):
        res = generator.feature_detection_labelling(array, wrong_config)
    config = [
        {"id": 0, "color": [10, 10, 200], "is_evaluate": True},
        {"id": 1, "color": [200, 10, 10], "is_evaluate": True},
        {"id": 2, "color": [10, 200, 10], "is_evaluate": True},
        {"id": 3, "color": [200, 200, 200], "is_evaluate": True},
    ]
    res = generator.feature_detection_labelling(array, config)
    assert res.shape == (array.shape[0], len(labels))
    assert res.tolist() == [
        [True, False, False, True],
        [True, True, True, False],
    ]


def test_feature_detection_labelling_sparse():
    """Test `feature_detection_labelling` function in `generator` module by
    considering a sparse labelling, *i.e.* the array contains unknown values
    (to mimic the non-evaluated label situations):
    * as a preliminary verification, check if passing string labels raises an
    AttributeError exception
    * test if label length is different from the list of values in the array
    * test if output shape is first input shape (batch size) + an additional
    dimension given by the `label_ids` length
    * test if both representation provides the same information (native array
    on the first hand and its one-hot version on the second hand)
    """
    array = np.array(
        [
            [
                [[10, 10, 200], [10, 10, 200], [10, 10, 200], [200, 10, 10]],
                [
                    [200, 200, 200],
                    [200, 200, 200],
                    [10, 10, 200],
                    [200, 10, 10],
                ],
                [
                    [200, 200, 200],
                    [200, 200, 200],
                    [200, 200, 200],
                    [10, 10, 200],
                ],
                [
                    [200, 200, 200],
                    [200, 200, 200],
                    [200, 200, 200],
                    [10, 10, 200],
                ],
            ],
            [
                [[200, 10, 10], [200, 10, 10], [10, 200, 10], [200, 10, 10]],
                [[200, 200, 200], [10, 200, 10], [10, 200, 10], [10, 200, 10]],
                [[200, 10, 10], [200, 10, 10], [200, 10, 10], [200, 200, 200]],
                [[200, 10, 10], [200, 10, 10], [10, 200, 10], [200, 200, 200]],
            ],
        ]
    )
    labels = np.unique(array.reshape(-1, 3), axis=0).tolist()[:-1]
    wrong_config = [
        {"id": "0", "color": [10, 10, 200], "is_evaluate": True},
        {"id": "1", "color": [200, 10, 10], "is_evaluate": True},
        {"id": "2", "color": [10, 200, 10], "is_evaluate": True},
    ]
    with pytest.raises(ValueError):
        res = generator.feature_detection_labelling(array, wrong_config)
    config = [
        {"id": 0, "color": [10, 10, 200], "is_evaluate": True},
        {"id": 1, "color": [200, 10, 10], "is_evaluate": True},
        {"id": 2, "color": [10, 200, 10], "is_evaluate": True},
    ]
    res = generator.feature_detection_labelling(array, config)
    assert len(labels) != np.amax(array) - np.amin(array) + 1
    assert res.tolist() == [[True, True, False], [False, True, True]]
    assert res.shape == (array.shape[0], len(labels))


def test_featdet_mapillary_generator(
    mapillary_image_size,
    mapillary_sample,
    mapillary_sample_config,
    nb_channels,
):
    """Test the data generator for the Mapillary dataset"""
    batch_size = 10
    config = utils.read_config(mapillary_sample_config)
    label_ids = [x["id"] for x in config["labels"]]
    gen = generator.create_generator(
        "mapillary",
        "featdet",
        mapillary_sample,
        mapillary_image_size,
        batch_size,
        config["labels"],
    )
    item = next(gen)
    assert len(item) == 2
    im_shape = item[0].shape
    assert im_shape == (
        batch_size,
        mapillary_image_size,
        mapillary_image_size,
        nb_channels,
    )
    label_shape = item[1].shape
    assert label_shape == (batch_size, len(label_ids))


def test_featdet_shape_generator(
    shapes_image_size, shapes_sample, shapes_sample_config, nb_channels
):
    """Test the data generator for the shape dataset"""
    batch_size = 10
    config = utils.read_config(shapes_sample_config)
    label_ids = [x["id"] for x in config["labels"]]
    gen = generator.create_generator(
        "shapes",
        "featdet",
        shapes_sample,
        shapes_image_size,
        batch_size,
        config["labels"],
    )
    item = next(gen)
    assert len(item) == 2
    im_shape = item[0].shape
    assert im_shape == (
        batch_size,
        shapes_image_size,
        shapes_image_size,
        nb_channels,
    )
    label_shape = item[1].shape
    assert label_shape == (batch_size, len(label_ids))


def test_semantic_segmentation_labelling_concise():
    """Test `semantic_segmentation_labelling` function in `generator` module by
    considering a concise labelling, *i.e.* the labels correspond to array
    values
    * as a preliminary verification, check if passing string labels raises an
    AttributeError exception
    * test if output shape is input shape + an additional dimension given by
    the `label_ids` length
    * test if both representation provides the same information (native array
    on the first hand and its one-hot version on the second hand)

    """
    array = np.array(
        [
            [
                [[200, 10, 10], [200, 10, 10], [200, 200, 200]],
                [[200, 200, 200], [200, 200, 200], [200, 10, 10]],
                [[200, 200, 200], [200, 200, 200], [200, 200, 200]],
            ],
            [
                [[200, 10, 10], [200, 10, 10], [10, 10, 200]],
                [[10, 200, 10], [10, 200, 10], [10, 10, 200]],
                [[200, 10, 10], [200, 10, 10], [10, 10, 200]],
            ],
        ]
    )
    labels = np.unique(array.reshape(-1, 3), axis=0).tolist()
    wrong_config = [
        {"id": "0", "color": [10, 10, 200], "is_evaluate": True},
        {"id": "1", "color": [200, 10, 10], "is_evaluate": True},
        {"id": "2", "color": [10, 200, 10], "is_evaluate": True},
        {"id": "3", "color": [200, 200, 200], "is_evaluate": True},
    ]
    with pytest.raises(ValueError):
        res = generator.semantic_segmentation_labelling(array, wrong_config)
    config = [
        {"id": 0, "color": [10, 10, 200], "is_evaluate": True},
        {"id": 1, "color": [200, 10, 10], "is_evaluate": True},
        {"id": 2, "color": [10, 200, 10], "is_evaluate": True},
        {"id": 3, "color": [200, 200, 200], "is_evaluate": True},
    ]
    res = generator.semantic_segmentation_labelling(array, config)
    assert res.shape == (array.shape[0], array.shape[1], array.shape[2], len(labels))
    assert res.tolist() == [
        [
            [
                [False, True, False, False],
                [False, True, False, False],
                [False, False, False, True],
            ],
            [
                [False, False, False, True],
                [False, False, False, True],
                [False, True, False, False],
            ],
            [
                [False, False, False, True],
                [False, False, False, True],
                [False, False, False, True],
            ],
        ],
        [
            [
                [False, True, False, False],
                [False, True, False, False],
                [True, False, False, False],
            ],
            [
                [False, False, True, False],
                [False, False, True, False],
                [True, False, False, False],
            ],
            [
                [False, True, False, False],
                [False, True, False, False],
                [True, False, False, False],
            ],
        ],
    ]


def test_semantic_segmentation_labelling_sparse():
    """Test `semantic_segmentation_labelling` function in `generator` module by
    considering a sparse labelling, *i.e.* the array contains unknown values
    (to mimic the non-evaluated label situations)
    * as a preliminary verification, check if passing string labels raises an
    AttributeError exception
    * test if output shape is input shape + an additional dimension given by
    the `label_ids` length
    * test if both representation provides the same information (native array
    on the first hand and its one-hot version on the second hand)

    """
    array = np.array(
        [
            [
                [[200, 10, 10], [200, 10, 10], [200, 200, 200]],
                [[200, 200, 200], [200, 200, 200], [200, 10, 10]],
                [[200, 200, 200], [100, 100, 100], [200, 200, 200]],
            ],
            [
                [[200, 10, 10], [200, 10, 10], [10, 10, 200]],
                [[200, 200, 200], [100, 100, 100], [10, 10, 200]],
                [[200, 10, 10], [200, 10, 10], [10, 10, 200]],
            ],
        ]
    )
    wrong_config = [
        {"id": "0", "color": [10, 10, 200], "is_evaluate": True},
        {"id": "2", "color": [10, 200, 10], "is_evaluate": True},
        {"id": "3", "color": [200, 200, 200], "is_evaluate": True},
    ]
    with pytest.raises(ValueError):
        res = generator.semantic_segmentation_labelling(array, wrong_config)
    config = [
        {"id": 0, "color": [10, 10, 200], "is_evaluate": True},
        {"id": 2, "color": [10, 200, 10], "is_evaluate": True},
        {"id": 3, "color": [200, 200, 200], "is_evaluate": True},
    ]
    labels = [item["id"] for item in config]
    res = generator.semantic_segmentation_labelling(array, config)
    assert len(labels) != np.amax(array) - np.amin(array) + 1
    assert res.shape == (array.shape[0], array.shape[1], array.shape[2], len(labels))
    assert res.tolist() == [
        [
            [
                [False, False, False],
                [False, False, False],
                [False, False, True],
            ],
            [
                [False, False, True],
                [False, False, True],
                [False, False, False],
            ],
            [
                [False, False, True],
                [False, False, False],
                [False, False, True],
            ],
        ],
        [
            [
                [False, False, False],
                [False, False, False],
                [True, False, False],
            ],
            [
                [False, False, True],
                [False, False, False],
                [True, False, False],
            ],
            [
                [False, False, False],
                [False, False, False],
                [True, False, False],
            ],
        ],
    ]


def test_semseg_mapillary_generator(
    mapillary_image_size,
    mapillary_sample,
    mapillary_sample_config,
    nb_channels,
):
    """Test the data generator for the Mapillary dataset"""
    batch_size = 10
    config = utils.read_config(mapillary_sample_config)
    label_ids = [x["id"] for x in config["labels"]]
    gen = generator.create_generator(
        "mapillary",
        "semseg",
        mapillary_sample,
        mapillary_image_size,
        batch_size,
        config["labels"],
    )
    item = next(gen)
    assert len(item) == 2
    im_shape = item[0].shape
    assert im_shape == (
        batch_size,
        mapillary_image_size,
        mapillary_image_size,
        nb_channels,
    )
    label_shape = item[1].shape
    assert label_shape == (
        batch_size,
        mapillary_image_size,
        mapillary_image_size,
        len(label_ids),
    )


def test_semseg_shape_generator(
    shapes_image_size, shapes_sample, shapes_sample_config, nb_channels
):
    """Test the data generator for the shape dataset"""
    batch_size = 10
    config = utils.read_config(shapes_sample_config)
    label_ids = [x["id"] for x in config["labels"]]
    gen = generator.create_generator(
        "shapes",
        "semseg",
        shapes_sample,
        shapes_image_size,
        batch_size,
        config["labels"],
    )
    item = next(gen)
    assert len(item) == 2
    im_shape = item[0].shape
    assert im_shape == (
        batch_size,
        shapes_image_size,
        shapes_image_size,
        nb_channels,
    )
    label_shape = item[1].shape
    assert label_shape == (
        batch_size,
        shapes_image_size,
        shapes_image_size,
        len(label_ids),
    )


def test_semseg_aerial_generator(
    aerial_image_size, aerial_sample, aerial_sample_config, nb_channels
):
    """Test the data generator for the AerialImage dataset"""
    batch_size = 4
    config = utils.read_config(aerial_sample_config)
    label_ids = [x["id"] for x in config["labels"]]
    gen = generator.create_generator(
        "aerial",
        "semseg",
        aerial_sample,
        aerial_image_size,
        batch_size,
        config["labels"],
    )
    item = next(gen)
    assert len(item) == 2
    im_shape = item[0].shape
    assert im_shape == (
        batch_size,
        aerial_image_size,
        aerial_image_size,
        nb_channels,
    )
    label_shape = item[1].shape
    assert label_shape == (
        batch_size,
        aerial_image_size,
        aerial_image_size,
        len(label_ids),
    )


def test_semseg_tanzania_generator(
    tanzania_image_size, tanzania_sample, tanzania_sample_config, nb_channels
):
    """Test the data generator for the Open AI Tanzania dataset"""
    batch_size = 3
    config = utils.read_config(tanzania_sample_config)
    label_ids = [x["id"] for x in config["labels"]]
    gen = generator.create_generator(
        "tanzania",
        "semseg",
        tanzania_sample,
        tanzania_image_size,
        batch_size,
        config["labels"],
    )
    item = next(gen)
    assert len(item) == 2
    im_shape = item[0].shape
    assert im_shape == (
        batch_size,
        tanzania_image_size,
        tanzania_image_size,
        nb_channels,
    )
    label_shape = item[1].shape
    assert label_shape == (
        batch_size,
        tanzania_image_size,
        tanzania_image_size,
        len(label_ids),
    )


def test_wrong_model_dataset_generator(shapes_sample_config):
    """Test a wrong model and wrong dataset"""
    dataset = "fake"
    model = "conquer_the_world"
    image_size = 10
    batch_size = 10
    datapath = "./tests/data/" + dataset + "/training"
    config = utils.read_config(shapes_sample_config)

    # wrong dataset name
    with pytest.raises(ValueError) as excinfo:
        generator.create_generator(
            dataset,
            "featdet",
            datapath,
            image_size,
            batch_size,
            config["labels"],
        )
    assert str(excinfo.value) == "Wrong dataset name {}".format(dataset)

    # wrong model name
    with pytest.raises(ValueError) as excinfo:
        generator.create_generator(
            "shapes", model, datapath, image_size, batch_size, config["labels"]
        )
    expected_failure_msg = "Wrong model name {} (choose amongst {})".format(
        model, AVAILABLE_MODELS
    )
    assert str(excinfo.value) == expected_failure_msg



"""Test config setup for training dataset handling
"""

import json
import math
import os

import pytest
from tensorflow.keras import backend as K


@pytest.fixture(autouse=True)
def clear_session_after_test():
    """Test wrapper to clean up after TensorFlow and CNTK tests.
    This wrapper runs for all the tests in the keras test suite.
    """
    yield
    if K.backend() == "tensorflow" or K.backend() == "cntk":
        K.clear_session()


@pytest.fixture
def nb_channels():
    return 3


@pytest.fixture
def kernel_size():
    return 3


@pytest.fixture
def conv_depth():
    return 8


@pytest.fixture
def conv_strides():
    return 2


@pytest.fixture
def pool_strides():
    return 2


@pytest.fixture
def pool_size():
    return 2


@pytest.fixture
def mapillary_image_size():
    return 224


@pytest.fixture
def mapillary_input_config():
    return "tests/data/mapillary/config_aggregate.json"


@pytest.fixture
def mapillary_sample_config():
    return "tests/data/mapillary/training.json"


@pytest.fixture
def mapillary_raw_sample():
    """Sample of Mapillary original images (called for populate a Mapillary
    dataset)

    """
    return "tests/data/mapillary/sample/"


@pytest.fixture
def mapillary_sample():
    """Sample of preprocessed Mapillary images"""
    return "tests/data/mapillary/training/"


@pytest.fixture
def mapillary_sample_without_labels_dir():
    return "tests/data/mapillary/sample_no_label/"


@pytest.fixture
def mapillary_nb_images(mapillary_sample):  # pylint: disable=W0621
    return len(os.listdir(os.path.join(mapillary_sample, "images")))


@pytest.fixture
def mapillary_nb_labels(mapillary_input_config):  # pylint: disable=W0621
    with open(mapillary_input_config) as fobj:
        config = json.load(fobj)
    return len(config["labels"])


@pytest.fixture(scope="session")
def mapillary_config(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("mapillary.json")


@pytest.fixture(scope="session")
def mapillary_temp_dir(tmpdir_factory):
    mapillary_subdir = tmpdir_factory.mktemp("mapillary", numbered=False)
    os.makedirs(os.path.join(mapillary_subdir, "images"))
    os.makedirs(os.path.join(mapillary_subdir, "labels"))
    os.makedirs(os.path.join(mapillary_subdir, "checkpoints"))
    return mapillary_subdir


@pytest.fixture
def shapes_image_size():
    return 64


@pytest.fixture
def shapes_nb_images():
    return 10


@pytest.fixture
def shapes_nb_labels():
    return 4


@pytest.fixture(scope="session")
def shapes_config(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("shapes.json")


@pytest.fixture
def shapes_sample_config():
    return "tests/data/shapes/training.json"


@pytest.fixture
def shapes_sample():
    return "tests/data/shapes/training/"


@pytest.fixture(scope="session")
def shapes_temp_dir(tmpdir_factory):
    shapes_subdir = tmpdir_factory.mktemp("shapes", numbered=False)
    os.makedirs(os.path.join(shapes_subdir, "images"))
    os.makedirs(os.path.join(shapes_subdir, "labels"))
    os.makedirs(os.path.join(shapes_subdir, "checkpoints"))
    return shapes_subdir


@pytest.fixture
def aerial_raw_image_size():
    return 5000


@pytest.fixture
def aerial_test_image_size():
    return 1000


@pytest.fixture
def aerial_image_size():
    return 240


@pytest.fixture
def aerial_nb_images():
    return 1


@pytest.fixture
def nb_tiles_per_image():
    return 10


@pytest.fixture
def aerial_nb_output_training_images():
    return 10


@pytest.fixture
def aerial_nb_output_testing_images(
    aerial_nb_images, aerial_test_image_size, aerial_image_size
):  # pylint: disable=W0621
    return aerial_nb_images * math.ceil(aerial_test_image_size / aerial_image_size) ** 2


@pytest.fixture
def aerial_nb_labels():
    return 2


@pytest.fixture(scope="session")
def aerial_training_config(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("aerial_training.json")


@pytest.fixture(scope="session")
def aerial_testing_config(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("aerial_testing.json")


@pytest.fixture
def aerial_sample_config():
    return "tests/data/aerial/preprocessed/240/training.json"


@pytest.fixture
def aerial_sample():
    return "tests/data/aerial/preprocessed/240/training/"


@pytest.fixture
def aerial_raw_sample():
    """Sample of AerialImage original images (called for populate a AerialImage
    dataset)

    """
    return "tests/data/aerial/input/training/"


@pytest.fixture(scope="session")
def aerial_training_temp_dir(tmpdir_factory):
    aerial_subdir = tmpdir_factory.mktemp("aerial_training", numbered=False)
    os.makedirs(os.path.join(aerial_subdir, "images"))
    os.makedirs(os.path.join(aerial_subdir, "labels"))
    os.makedirs(os.path.join(aerial_subdir, "checkpoints"))
    return aerial_subdir


@pytest.fixture(scope="session")
def aerial_testing_temp_dir(tmpdir_factory):
    aerial_subdir = tmpdir_factory.mktemp("aerial_testing", numbered=False)
    os.makedirs(os.path.join(aerial_subdir, "images"))
    os.makedirs(os.path.join(aerial_subdir, "labels"))
    os.makedirs(os.path.join(aerial_subdir, "checkpoints"))
    return aerial_subdir


@pytest.fixture(scope="session")
def aerial_temp_conf(tmpdir_factory):
    temp_conf = tmpdir_factory.mktemp(".").join("aerial.json")
    return temp_conf


@pytest.fixture
def tanzania_raw_image_size():
    return 1000


@pytest.fixture
def tanzania_image_size():
    return 384


@pytest.fixture
def tanzania_nb_images():
    return 1


@pytest.fixture
def tanzania_nb_output_training_images():
    return 10


@pytest.fixture
def tanzania_nb_output_testing_images(
    tanzania_nb_images, tanzania_raw_image_size, tanzania_image_size
):  # pylint: disable=W0621
    return (
        tanzania_nb_images
        * math.ceil(tanzania_raw_image_size / tanzania_image_size) ** 2
    )


@pytest.fixture
def tanzania_nb_labels():
    return 4


@pytest.fixture(scope="session")
def tanzania_training_config(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("tanzania_training.json")


@pytest.fixture(scope="session")
def tanzania_testing_config(tmpdir_factory):
    return tmpdir_factory.getbasetemp().join("tanzania_testing.json")


@pytest.fixture
def tanzania_sample_config():
    return "tests/data/tanzania/preprocessed/384/training.json"


@pytest.fixture
def tanzania_sample():
    return "tests/data/tanzania/preprocessed/384/training/"


@pytest.fixture
def tanzania_raw_sample():
    """Sample of Tanzania original images (called for populate a
    TanzaniaDataset)

    """
    return "tests/data/tanzania/input/training/"


@pytest.fixture
def tanzania_example_image(tanzania_raw_sample):  # pylint: disable=W0621
    return os.path.join(tanzania_raw_sample, "images", "tanzania_sample.tif")


@pytest.fixture
def tanzania_example_labels(tanzania_raw_sample):  # pylint: disable=W0621
    return os.path.join(tanzania_raw_sample, "labels", "tanzania_sample.geojson")


@pytest.fixture(scope="session")
def tanzania_training_temp_dir(tmpdir_factory):
    tanzania_subdir = tmpdir_factory.mktemp("tanzania_training", numbered=False)
    os.makedirs(os.path.join(tanzania_subdir, "images"))
    os.makedirs(os.path.join(tanzania_subdir, "labels"))
    os.makedirs(os.path.join(tanzania_subdir, "checkpoints"))
    return tanzania_subdir


@pytest.fixture(scope="session")
def tanzania_testing_temp_dir(tmpdir_factory):
    tanzania_subdir = tmpdir_factory.mktemp("tanzania_testing", numbered=False)
    os.makedirs(os.path.join(tanzania_subdir, "images"))
    os.makedirs(os.path.join(tanzania_subdir, "labels"))
    os.makedirs(os.path.join(tanzania_subdir, "checkpoints"))
    return tanzania_subdir


@pytest.fixture(scope="session")
def tanzania_temp_conf(tmpdir_factory):
    temp_conf = tmpdir_factory.mktemp(".").join("tanzania.json")
    return temp_conf


@pytest.fixture(scope="session")
def datapath_repo(tmpdir_factory):
    datapath_subdir = tmpdir_factory.mktemp("datapath", numbered=False)
    return datapath_subdir



"""Unit tests that address geometries.py module functions
"""

import geopandas as gpd
import numpy as np
import pytest
from osgeo import gdal
from shapely.geometry import MultiPolygon, Polygon

from deeposlandia.geometries import (
    convert_to_geocoord,
    extract_geometry_vertices,
    extract_points_from_polygon,
    extract_tile_items,
    get_geocoord,
    get_image_features,
    get_pixel,
    get_tile_footprint,
    pixel_to_geocoord,
    rasterize_polygons,
    retrieve_area_color,
    vectorize_mask,
)


def test_get_pixel():
    """Test the transformation from georeferenced coordinates to pixel.

    The transformation conserves the coordinate data structure (scalar or
    list).
    """
    geocoords = [10000.0, 15000.0, 20000.0]
    min_coord, max_coord = 0, 30000.0
    size_in_pixel = 500
    true_pixel_coord = [166, 250, 333]
    pixel_coord = get_pixel(geocoords, min_coord, max_coord, size_in_pixel)
    assert np.all(pixel_coord == true_pixel_coord)
    single_geocoord = geocoords[0]
    single_pixel_coord = get_pixel(single_geocoord, min_coord, max_coord, size_in_pixel)
    assert single_pixel_coord == true_pixel_coord[0]
    str_geocoord = "15000"
    with pytest.raises(TypeError):
        get_pixel(str_geocoord, min_coord, max_coord, size_in_pixel)


def test_get_geocoord():
    """Test the transformation from pixel to georeferenced coordinates.

    The transformation conserves the coordinate data structure (scalar or
    list).
    """
    pixels = [166, 250, 333]
    min_coord, max_coord = 0, 30000.0
    size_in_pixel = 500
    true_geocoord = [9960.0, 15000.0, 19980.0]
    geocoord = get_geocoord(pixels, min_coord, max_coord, size_in_pixel)
    assert np.all(geocoord == true_geocoord)
    single_pixel = pixels[0]
    single_geocoord = get_geocoord(single_pixel, min_coord, max_coord, size_in_pixel)
    assert single_geocoord == true_geocoord[0]
    str_pixel = "250"
    with pytest.raises(TypeError):
        get_geocoord(str_pixel, min_coord, max_coord, size_in_pixel)


def test_get_image_features(tanzania_example_image):
    """Test the image geographic feature recovering:
    - 'south', 'north', 'west' and 'east' are the image geographic coordinates,
    hence floating numbers
    - west is smaller than east
    - south is smaller than north
    - srid is an integer geocode
    - width and height are strictly positive int, as they represent the image
    size, in pixels
    """
    dataset = gdal.Open(str(tanzania_example_image))
    geofeatures = get_image_features(dataset)
    assert isinstance(geofeatures["south"], float)
    assert isinstance(geofeatures["north"], float)
    assert isinstance(geofeatures["east"], float)
    assert isinstance(geofeatures["west"], float)
    assert geofeatures["west"] < geofeatures["east"]
    assert geofeatures["south"] < geofeatures["north"]
    assert isinstance(geofeatures["srid"], int)
    assert isinstance(geofeatures["width"], int)
    assert isinstance(geofeatures["height"], int)
    assert geofeatures["width"] > 0
    assert geofeatures["height"] > 0


def test_extract_points_from_polygon(tanzania_example_image):
    """Test a polygon point extraction.

    Within a 1000x1000 pixel original image, consider 500x500 tiles, and more
    specifically the right-bottom tile. One wants to retrieve a triangle
    whose coordinates are as follows:
    - (image_width/2, image_height/2)
    - (image_width/2, image_height)
    - (image_width*3/4, image_height/2)

    The point coordinate representation must be inverted between georeferenced
    points and 2D-'numpy.array' pixel points: in the latter, the first
    (resp. the second) dimension corresponds to rows (resp.columns).
    """
    dataset = gdal.Open(str(tanzania_example_image))
    geofeatures = get_image_features(dataset)
    min_x = min_y = 500
    x1 = geofeatures["west"] + (geofeatures["east"] - geofeatures["west"]) / 2
    y1 = geofeatures["south"] + (geofeatures["north"] - geofeatures["south"]) / 2
    x2 = x1 + (geofeatures["east"] - x1) / 2
    y2 = geofeatures["south"]
    polygon = Polygon(((x1, y1), (x1, y2), (x2, y1), (x1, y1)))
    points = extract_points_from_polygon(polygon, geofeatures, min_x, min_y)
    expected_points = np.array([[0, 0], [500, 0], [0, 250], [0, 0]])
    assert np.all(points == expected_points)


def test_square_tile_footprint(tanzania_example_image):
    """Test a tile footprint recovery, based on the reference test image (see
    'tests/data/tanzania/input/training/').

    The full image is considered as the tile, its bounds must equal the image
    coordinates.
    """
    dataset = gdal.Open(str(tanzania_example_image))
    geofeatures = get_image_features(dataset)
    min_x = min_y = 0
    tile_width = dataset.RasterXSize
    tile_footprint = get_tile_footprint(geofeatures, min_x, min_y, tile_width)
    assert tile_footprint.is_valid
    tile_bounds = tile_footprint.bounds
    assert geofeatures["north"] in tile_bounds
    assert geofeatures["south"] in tile_bounds
    assert geofeatures["east"] in tile_bounds
    assert geofeatures["west"] in tile_bounds


def test_rectangle_tile_footprint(tanzania_example_image):
    """Test a tile footprint recovery, based on the reference test image (see
    'tests/data/tanzania/input/training/').

    The considered tile is the top-half of the image, its bounds must equal
    the image coordinates, except the south bound that must equal the mean
    between north and south coordinates.
    """
    dataset = gdal.Open(str(tanzania_example_image))
    geofeatures = get_image_features(dataset)
    min_x = min_y = 0
    tile_width = dataset.RasterXSize
    tile_height = int(dataset.RasterYSize / 2)
    tile_footprint = get_tile_footprint(
        geofeatures, min_x, min_y, tile_width, tile_height
    )
    assert tile_footprint.is_valid
    tile_bounds = tile_footprint.bounds
    tile_south = (
        geofeatures["south"] + (geofeatures["north"] - geofeatures["south"]) / 2
    )
    assert tile_south in tile_bounds
    assert geofeatures["north"] in tile_bounds
    assert geofeatures["east"] in tile_bounds
    assert geofeatures["west"] in tile_bounds


def test_extract_empty_tile_items(tanzania_example_image, tanzania_example_labels):
    """Test the extraction of polygons that overlap a given squared tile, based
    on a reference test image (see 'tests/data/tanzania/input/training/').

    The tests is focused on an empty tile, that must provide an empty item set.
    """
    dataset = gdal.Open(str(tanzania_example_image))
    geofeatures = get_image_features(dataset)
    labels = gpd.read_file(tanzania_example_labels)
    labels = labels.loc[~labels.geometry.isna(), ["condition", "geometry"]]
    none_mask = [lc is None for lc in labels.condition]
    labels.loc[none_mask, "condition"] = "Complete"
    empty_tile_items = extract_tile_items(geofeatures, labels, 450, 450, 100, 100)
    assert empty_tile_items.shape[0] == 0


def test_extract_tile_items(tanzania_example_image, tanzania_example_labels):
    """Test the extraction of polygons that overlap a given squared tile, based
    on a reference test image (see 'tests/data/tanzania/input/training/').

    The tests check that:
    - the example image contains 7 valid items
    - the items are 'Polygon' (in opposition to 'MultiPolygon')
    - the item union is contained into the tile footprint (overlapping items
    are cutted out so as out-of-image parts are removed)
    """
    dataset = gdal.Open(str(tanzania_example_image))
    geofeatures = get_image_features(dataset)
    labels = gpd.read_file(tanzania_example_labels)
    labels = labels.loc[~labels.geometry.isna(), ["condition", "geometry"]]
    none_mask = [lc is None for lc in labels.condition]
    labels.loc[none_mask, "condition"] = "Complete"
    tile_items = extract_tile_items(geofeatures, labels, 0, 0, 1000, 1000)
    expected_items = 7
    assert tile_items.shape[0] == expected_items
    assert np.all([geom.is_valid for geom in tile_items["geometry"]])
    assert np.all([geom.geom_type == "Polygon" for geom in tile_items["geometry"]])
    item_bounds = tile_items.unary_union.bounds
    assert (
        item_bounds[0] >= geofeatures["west"] and item_bounds[0] <= geofeatures["east"]
    )
    assert (
        item_bounds[1] >= geofeatures["south"]
        and item_bounds[1] <= geofeatures["north"]
    )
    assert (
        item_bounds[2] >= geofeatures["west"] and item_bounds[2] <= geofeatures["east"]
    )
    assert (
        item_bounds[3] >= geofeatures["south"]
        and item_bounds[3] <= geofeatures["north"]
    )


def test_extract_geometry_vertices(tanzania_raw_image_size):
    """Test the polygon vertice extraction from a raster mask

    Test a simple case with two expected polygons. The considered function uses
    OpenCV library to find polygon contours, and some approximated methods are
    implied. Hence exact polygon vertice coordinates are not tested.

    Two variant are tested, regarding the "structure" parameter, fundamental in
    OpenCV operations. If it is too large compared to polygon sizes, they are
    not detected.
    """
    mask = np.zeros([10, 10], dtype=np.uint8)
    x1, y1 = 1, 2
    x2, y2 = 3, 4
    x3, y3 = 3, 6
    x4, y4 = 7, 9
    mask[y1:y2, x1:x2] = 1
    mask[y3:y4, x3:x4] = 1
    undetected_vertices, _ = extract_geometry_vertices(mask, structure_size=(10, 10))
    assert len(undetected_vertices) == 0
    bigger_mask = np.zeros(
        [tanzania_raw_image_size, tanzania_raw_image_size], dtype=np.uint8
    )
    x1, y1 = 100, 200
    x2, y2 = 300, 400
    x3, y3 = 300, 600
    x4, y4 = 700, 900
    bigger_mask[y1:y2, x1:x2] = 1
    bigger_mask[y3:y4, x3:x4] = 1
    polygon_vertices, _ = extract_geometry_vertices(
        bigger_mask, structure_size=(10, 10)
    )
    assert len(polygon_vertices) == 2


def test_retrieve_area_color(tanzania_raw_image_size):
    """Test the label retrieving function, in order to assign a label to a
    given area thanks to its color

    One tests three cases:
      - an empty area must return "0" label
      - an area filled with a color must return corresponding label
      - an area partially filled must return the most encountered label
    """
    label_dicts = [{"id": 0, "color": [0, 0, 0]}, {"id": 1, "color": [50, 200, 50]}]
    data = np.zeros(
        [tanzania_raw_image_size, tanzania_raw_image_size, 3], dtype=np.uint8
    )
    x1, y1 = 100, 200
    x2, y2 = 300, 400
    contour = np.array(
        [[[x1, y1]], [[x1, y2]], [[x2, y2]], [[x2, y1]]],
    )
    assert retrieve_area_color(data, contour, label_dicts) == 0
    data[y1:y2, x1:x2] = label_dicts[1]["color"]
    assert retrieve_area_color(data, contour, label_dicts) == 1
    x2 += 100
    contour = np.array(
        [[[x1, y1]], [[x1, y2]], [[x2, y2]], [[x2, y1]]],
    )
    assert retrieve_area_color(data, contour, label_dicts) == 1


def test_vectorize_mask(tanzania_raw_image_size):
    """Test the mask vectorization operation, that transform raster mask into a
    MultiPolygon.

    Test a simple case with two expected polygons. The considered function uses
    OpenCV library to find polygon contours, and some approximated methods are
    implied. Hence exact polygon vertice coordinates are not tested.
    """
    label_dicts = [{"id": 1, "color": [50, 200, 50]}, {"id": 2, "color": [200, 50, 50]}]
    mask = np.zeros([tanzania_raw_image_size, tanzania_raw_image_size], dtype=np.uint8)
    data = np.zeros(
        [tanzania_raw_image_size, tanzania_raw_image_size, 3], dtype=np.uint8
    )
    empty_labels, empty_multipolygon = vectorize_mask(mask, data, label_dicts)
    assert len(empty_labels) == 0
    assert len(empty_multipolygon) == 0
    x1, y1 = 100, 200
    x2, y2 = 300, 400
    x3, y3 = 300, 600
    x4, y4 = 700, 900
    mask[y1:y2, x1:x2] = label_dicts[0]["id"]
    mask[y3:y4, x3:x4] = label_dicts[1]["id"]
    data[y1:y2, x1:x2] = label_dicts[0]["color"]
    data[y3:y4, x3:x4] = label_dicts[1]["color"]
    labels, multipolygon = vectorize_mask(mask, data, label_dicts)
    assert len(labels) == 2
    assert np.sum(labels == 1) == 1
    assert np.sum(labels == 2) == 1
    assert len(multipolygon) == 2


def test_rasterize_polygons(tanzania_raw_image_size):
    """Test the rasterization process

    Considering the "no-polygon" case, the function must return an empty mask.

    Considering a polygon that fill the left part of the original image. The
    rasterized mask must be filled with "1" on this part, and with "0" on the
    right part.
    """
    mask = rasterize_polygons(
        [], np.array([]), tanzania_raw_image_size, tanzania_raw_image_size
    )
    assert mask.shape == (tanzania_raw_image_size, tanzania_raw_image_size)
    assert np.unique(mask) == np.array([0])
    x1 = int(tanzania_raw_image_size / 3)
    x2 = int(2 * tanzania_raw_image_size / 3)
    polygon1 = Polygon(
        shell=(
            (0, 0),
            (x1, 0),
            (x1, tanzania_raw_image_size),
            (0, tanzania_raw_image_size),
            (0, 0),
        )
    )
    polygon2 = Polygon(
        shell=(
            (x1, 0),
            (x2, 0),
            (x2, tanzania_raw_image_size),
            (x1, tanzania_raw_image_size),
            (x1, 0),
        )
    )
    labels = [1, 2]
    mask = rasterize_polygons(
        MultiPolygon([polygon1, polygon2]),
        np.array(labels),
        tanzania_raw_image_size,
        tanzania_raw_image_size,
    )
    mask_polygon_1 = mask[:tanzania_raw_image_size, :x1]
    assert np.unique(mask_polygon_1) == labels[0]
    mask_polygon_2 = mask[:tanzania_raw_image_size, (1 + x1) : x2]
    assert np.unique(mask_polygon_2) == labels[1]
    mask_no_polygon = mask[:tanzania_raw_image_size, (1 + x2) :]
    assert np.unique(mask_no_polygon) == 0


def test_pixel_to_geocoord(tanzania_example_image, tanzania_raw_image_size):
    """Test the transformation of a Polygon from pixel to georeferenced
    coordinates

    Use the full image footprint as a reference polygon.
    """
    dataset = gdal.Open(str(tanzania_example_image))
    geofeatures = get_image_features(dataset)
    polygon = Polygon(
        shell=(
            (0, 0),
            (tanzania_raw_image_size, 0),
            (tanzania_raw_image_size, tanzania_raw_image_size),
            (0, tanzania_raw_image_size),
            (0, 0),
        )
    )
    expected_points = np.array(
        [
            [geofeatures["west"], geofeatures["north"]],
            [geofeatures["east"], geofeatures["north"]],
            [geofeatures["east"], geofeatures["south"]],
            [geofeatures["west"], geofeatures["south"]],
            [geofeatures["west"], geofeatures["north"]],
        ]
    )
    points = pixel_to_geocoord(polygon.exterior, geofeatures)
    assert np.all(points == expected_points)


def test_convert_to_geocoord(tanzania_example_image, tanzania_raw_image_size):
    """Test the convertion of a set of pixel-referenced polygons to
    georeferenced ones.

    Some of the polygon may include holes (hence interior points). We test the
    following design, where there are two polygons, of whom one has a hole:
        ____
       |1110|
       |1010|
       |1110|
       |0002|
        ----
    """
    x0 = y0 = 0
    x1 = y1 = int(tanzania_raw_image_size / 4)
    x2 = y2 = int(tanzania_raw_image_size / 2)
    x3 = y3 = int(tanzania_raw_image_size * 3 / 4)
    x4 = y4 = tanzania_raw_image_size
    polygon1 = Polygon(
        shell=((x0, y0), (x3, y0), (x3, y3), (x0, y3), (x0, y0)),
        holes=[((x1, y1), (x2, y1), (x2, y2), (x1, y2), (x1, y1))],
    )
    polygon2 = Polygon(shell=((x3, y3), (x4, y3), (x4, y4), (x3, y4), (x3, y3)))
    multipolygon = MultiPolygon([polygon1, polygon2])
    dataset = gdal.Open(str(tanzania_example_image))
    geofeatures = get_image_features(dataset)
    converted_multipolygon = convert_to_geocoord(multipolygon, geofeatures)
    expected_x = [
        (geofeatures["west"] + (geofeatures["east"] - geofeatures["west"]) * i)
        for i in np.linspace(0, 1, 5)
    ]
    expected_y = [
        (geofeatures["north"] + (geofeatures["south"] - geofeatures["north"]) * i)
        for i in np.linspace(0, 1, 5)
    ]
    expected_polygon1 = Polygon(
        shell=(
            (expected_x[0], expected_y[0]),
            (expected_x[3], expected_y[0]),
            (expected_x[3], expected_y[3]),
            (expected_x[0], expected_y[3]),
            (expected_x[0], expected_y[0]),
        ),
        holes=[
            (
                (expected_x[1], expected_y[1]),
                (expected_x[2], expected_y[1]),
                (expected_x[2], expected_y[2]),
                (expected_x[1], expected_y[2]),
                (expected_x[1], expected_y[1]),
            )
        ],
    )
    expected_polygon2 = Polygon(
        shell=(
            (expected_x[3], expected_y[3]),
            (expected_x[4], expected_y[3]),
            (expected_x[4], expected_y[4]),
            (expected_x[3], expected_y[4]),
            (expected_x[3], expected_y[3]),
        )
    )
    assert converted_multipolygon[0] == expected_polygon1
    assert converted_multipolygon[1] == expected_polygon2



"""Unit test related to the simple layer creation
"""

from tensorflow import keras as K

from deeposlandia.network import ConvolutionalNeuralNetwork


def test_convolution_shape(shapes_image_size, kernel_size, conv_depth, conv_strides):
    """Test the convolution operation through its output layer shape"""
    cnn = ConvolutionalNeuralNetwork("test", shapes_image_size)
    y = cnn.convolution(
        cnn.in_tensor,
        nb_filters=conv_depth,
        kernel_size=kernel_size,
        strides=conv_strides,
        block_name="convtest",
    )
    model = K.models.Model(cnn.in_tensor, y)
    output_shape = model.output_shape
    assert len(output_shape) == 4
    assert output_shape[1:] == (
        shapes_image_size // conv_strides,
        shapes_image_size // conv_strides,
        conv_depth,
    )


def test_transposed_convolution_shape(
    shapes_image_size, conv_depth, kernel_size, conv_strides
):
    """Test the transposed convolution operation through its output layer shape"""
    cnn = ConvolutionalNeuralNetwork("test", shapes_image_size)
    y = cnn.transposed_convolution(
        cnn.in_tensor,
        nb_filters=conv_depth,
        kernel_size=kernel_size,
        strides=conv_strides,
        block_name="transconvtest",
    )
    model = K.models.Model(cnn.in_tensor, y)
    output_shape = model.output_shape
    assert len(output_shape) == 4
    assert output_shape[1:] == (
        shapes_image_size * conv_strides,
        shapes_image_size * conv_strides,
        conv_depth,
    )


def test_maxpooling_shape(shapes_image_size, nb_channels, pool_size, pool_strides):
    """Test the max pooling operation through its output layer shape"""
    cnn = ConvolutionalNeuralNetwork("test", shapes_image_size, nb_channels)
    y = cnn.maxpool(
        cnn.in_tensor, pool_size=pool_size, strides=pool_strides, block_name="pooltest"
    )
    model = K.models.Model(cnn.in_tensor, y)
    output_shape = model.output_shape
    assert len(output_shape) == 4
    assert output_shape[1:] == (
        shapes_image_size // pool_strides,
        shapes_image_size // pool_strides,
        nb_channels,
    )


def test_dense_shape(shapes_image_size, conv_depth):
    """Test the fully-connected layer through its output shape"""
    cnn = ConvolutionalNeuralNetwork("test", shapes_image_size)
    y = cnn.dense(cnn.in_tensor, depth=conv_depth, block_name="fctest")
    model = K.models.Model(cnn.in_tensor, y)
    output_shape = model.output_shape
    assert len(output_shape) == 4
    assert output_shape[1:] == (
        shapes_image_size,
        shapes_image_size,
        conv_depth,
    )


def test_flatten_shape(shapes_image_size, nb_channels):
    """Test the flattening layer through its output shape"""
    cnn = ConvolutionalNeuralNetwork(
        "test", image_size=shapes_image_size, nb_channels=nb_channels
    )
    y = cnn.flatten(cnn.in_tensor, block_name="flattentest")
    model = K.models.Model(cnn.in_tensor, y)
    output_shape = model.output_shape
    assert len(output_shape) == 2
    assert output_shape[1] == shapes_image_size * shapes_image_size * nb_channels


def test_layer_name(shapes_image_size, kernel_size, conv_depth, conv_strides):
    """Test the convolution operation through its output layer shape"""
    cnn = ConvolutionalNeuralNetwork("test", shapes_image_size)
    y = cnn.convolution(
        cnn.in_tensor,
        nb_filters=conv_depth,
        kernel_size=kernel_size,
        strides=conv_strides,
    )
    y = cnn.convolution(
        y, nb_filters=conv_depth, kernel_size=kernel_size, strides=conv_strides
    )
    model = K.models.Model(cnn.in_tensor, y)
    assert [layer.name for layer in model.layers[1:]] == [
        "conv2d",
        "batch_normalization",
        "activation",
        "conv2d_1",
        "batch_normalization_1",
        "activation_1",
    ]



"""Unit test related to data folder organization
"""

import os

from deeposlandia.utils import (
    prepare_input_folder,
    prepare_output_folder,
    prepare_preprocessed_folder,
)


def test_input_folder(datapath_repo):
    """Test the existence of the raw dataset directory when using
    ̀utils.prepare_input_folder`

    """
    datapath = str(datapath_repo)
    dataset = "shapes"
    prepare_input_folder(datapath, dataset)
    assert os.path.isdir(os.path.join(datapath, dataset, "input"))


def test_preprocessed_folder(datapath_repo):
    """Test the creation of the preprocessed data repositories, by checking the
    full expected tree, *i.e.* considering training, validation and testing
    repositories within an instance-specific folder, and images and labels
    repositories wihtin each of these subrepositories

    """
    datapath = str(datapath_repo)
    dataset = "shapes"
    image_size = 64
    prepare_preprocessed_folder(datapath, dataset, image_size)
    assert os.path.isdir(os.path.join(datapath, dataset, "preprocessed"))
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
        )
    )
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
            "training",
        )
    )
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
            "training",
            "images",
        )
    )
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
            "training",
            "labels",
        )
    )
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
            "validation",
        )
    )
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
            "validation",
            "images",
        )
    )
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
            "validation",
            "labels",
        )
    )
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
            "testing",
        )
    )
    assert os.path.isdir(
        os.path.join(
            datapath,
            dataset,
            "preprocessed",
            str(image_size),
            "testing",
            "images",
        )
    )


def test_output_folder(datapath_repo):
    """Test the creation of the dataset output repository, by checking the full
    expected tree, *i.e.* until the instance-specific checkpoint folder

    """
    datapath = str(datapath_repo)
    dataset = "shapes"
    image_size = 100
    model = "feature_detection"
    output_folder = prepare_output_folder(datapath, dataset, image_size, model)
    assert len(output_folder.keys()) == 6
    assert os.path.isdir(
        os.path.join(datapath, dataset, "output", model, "checkpoints")
    )
    assert os.path.isdir(
        os.path.join(datapath, dataset, "output", model, "predicted_labels")
    )
    dataset = "aerial"
    prepare_output_folder(datapath, dataset, image_size, model)
    assert os.path.isdir(
        os.path.join(datapath, dataset, "output", model, "predicted_geometries")
    )
    assert os.path.isdir(
        os.path.join(datapath, dataset, "output", model, "predicted_rasters")
    )



"""Unit test related to the feature detection model
"""

from tensorflow import keras as K

from deeposlandia.feature_detection import FeatureDetectionNetwork


def test_simple_network_architecture(shapes_image_size, nb_channels, shapes_nb_labels):
    """Test a simple feature detection network"""
    net = FeatureDetectionNetwork(
        "fd_test",
        image_size=shapes_image_size,
        nb_channels=nb_channels,
        nb_labels=shapes_nb_labels,
    )
    model = K.models.Model(net.in_tensor, net.out_tensor)
    input_shape = model.input_shape
    output_shape = model.output_shape
    assert len(input_shape) == 4
    assert input_shape[1:] == (
        shapes_image_size,
        shapes_image_size,
        nb_channels,
    )
    assert len(output_shape) == 2
    assert output_shape[1] == shapes_nb_labels


def test_vgg16_network_architecture(
    mapillary_image_size, nb_channels, mapillary_nb_labels
):
    """Test a VGG16-inspired feature detection network"""
    net = FeatureDetectionNetwork(
        "fd_test",
        image_size=mapillary_image_size,
        nb_channels=nb_channels,
        nb_labels=mapillary_nb_labels,
        architecture="vgg",
    )
    model = K.models.Model(net.in_tensor, net.out_tensor)
    input_shape = model.input_shape
    output_shape = model.output_shape
    assert len(input_shape) == 4
    assert input_shape[1:] == (
        mapillary_image_size,
        mapillary_image_size,
        nb_channels,
    )
    assert len(output_shape) == 2
    assert output_shape[1] == mapillary_nb_labels


def test_inception_network_architecture(
    mapillary_image_size, nb_channels, mapillary_nb_labels
):
    """Test a Inception-inspired feature detection network"""
    net = FeatureDetectionNetwork(
        "fd_test",
        image_size=mapillary_image_size,
        nb_channels=nb_channels,
        nb_labels=mapillary_nb_labels,
        architecture="inception",
    )
    model = K.models.Model(net.in_tensor, net.out_tensor)
    input_shape = model.input_shape
    output_shape = model.output_shape
    assert len(input_shape) == 4
    assert input_shape[1:] == (
        mapillary_image_size,
        mapillary_image_size,
        nb_channels,
    )
    assert len(output_shape) == 2
    assert output_shape[1] == mapillary_nb_labels



"""Unit test related to model training with Keras API
"""

import os

import numpy as np
from tensorflow import keras as K

from deeposlandia.feature_detection import FeatureDetectionNetwork
from deeposlandia.generator import create_generator
from deeposlandia.utils import read_config


def test_model_training(
    shapes_image_size,
    shapes_sample,
    shapes_sample_config,
    shapes_temp_dir,
    shapes_nb_images,
):
    """Test the training of a simple neural network with Keras API, as well as
    model inference and trained model backup

    One big test function to avoid duplicating the training operations (that
    can be long)
    """
    batch_size = 10
    nb_epochs = 1
    nb_steps = shapes_nb_images // batch_size
    config = read_config(shapes_sample_config)
    label_ids = [x["id"] for x in config["labels"] if x["is_evaluate"]]
    gen = create_generator(
        "shapes",
        "featdet",
        shapes_sample,
        shapes_image_size,
        batch_size,
        config["labels"],
    )
    cnn = FeatureDetectionNetwork(
        "test", image_size=shapes_image_size, nb_labels=len(label_ids)
    )
    model = K.models.Model(cnn.in_tensor, cnn.out_tensor)
    model.compile(loss="binary_crossentropy", optimizer="adam", metrics=["acc"])
    hist = model.fit_generator(gen, epochs=nb_epochs, steps_per_epoch=nb_steps)
    assert len(hist.history) == 2
    assert all(k in hist.history.keys() for k in ["acc", "loss"])
    assert hist.history["acc"][0] >= 0 and hist.history["acc"][0] <= 1

    test_image = np.random.randint(
        0, 255, [batch_size, shapes_image_size, shapes_image_size, 3]
    )
    score = model.predict(test_image)
    assert score.shape == (batch_size, len(label_ids))
    assert all(0 <= s <= 1 for s in score.ravel())

    backup_filename = os.path.join(
        str(shapes_temp_dir),
        "checkpoints",
        "test_model_{:02d}.h5".format(nb_epochs),
    )
    model.save(backup_filename)
    assert os.path.isfile(backup_filename)


def test_model_backup_loading(shapes_image_size, shapes_sample_config, shapes_temp_dir):
    """Test the model checkpoint recovering"""
    config = read_config(shapes_sample_config)
    label_ids = [x["id"] for x in config["labels"] if x["is_evaluate"]]

    cnn = FeatureDetectionNetwork(
        "test", image_size=shapes_image_size, nb_labels=len(label_ids)
    )
    model = K.models.Model(cnn.in_tensor, cnn.out_tensor)
    old_weights = model.get_weights()
    checkpoint_path = os.path.join(str(shapes_temp_dir), "checkpoints")
    if os.path.isdir(checkpoint_path):
        checkpoints = os.listdir(checkpoint_path)
        if len(checkpoints) > 0:
            model_checkpoint = max(checkpoints)
            trained_model_epoch = int(model_checkpoint[-5:-3])
            checkpoint_complete_path = os.path.join(checkpoint_path, model_checkpoint)
            model.load_weights(checkpoint_complete_path)
        else:
            trained_model_epoch = 0
    else:
        trained_model_epoch = 0
    new_weights = model.get_weights()
    assert trained_model_epoch > 0
    assert len(old_weights) == len(new_weights)
    assert old_weights[0].shape == new_weights[0].shape
    # Test if old and new weights are different (at least for one layer)
    assert any(not np.allclose(lhs, rhs) for lhs, rhs in zip(old_weights, new_weights))



"""Unit tests dedicated to geodataset predictions
"""

import os
from pathlib import Path

import numpy as np
import pytest
from osgeo import gdal

from deeposlandia import geoinference


def test_get_image_paths(tanzania_image_size):
    """Test the image path getting function

    Preprocessed image filenames must end with ".png"
    """
    filenames = geoinference.get_image_paths(
        f"./tests/data/tanzania/preprocessed/{tanzania_image_size}/testing/",
        "tanzania_sample",
    )
    assert np.all([f.endswith(".png") for f in filenames])


def test_extract_images(tanzania_image_size, tanzania_nb_output_testing_images):
    """Test the image extraction function, that retrieve the accurate data in a
    'numpy.array' starting from a list of filenames

    This image data must be shaped as (nb_filenames, image_size, image_size,
    3).
    """
    filenames = geoinference.get_image_paths(
        f"./tests/data/tanzania/preprocessed/{tanzania_image_size}/testing/",
        "tanzania_sample",
    )
    images = geoinference.extract_images(filenames)
    assert len(images.shape) == 4
    assert images.shape[0] == tanzania_nb_output_testing_images
    assert images.shape[1] == tanzania_image_size
    assert images.shape[2] == tanzania_image_size
    assert images.shape[3] == 3


def test_get_labels(tanzania_image_size, tanzania_nb_labels):
    """Test the label retrieving from dataset glossary"""
    labels = geoinference.get_labels("./tests/data", "tanzania", tanzania_image_size)
    assert len(labels) == tanzania_nb_labels
    with pytest.raises(ValueError):
        geoinference.get_labels(
            "./elsewhere-on-the-disk", "wrong-dataset", tanzania_image_size
        )


def test_get_trained_model(tanzania_image_size, tanzania_nb_labels):
    """Test the semantic segmentation model retrieving

    Geoinference implies getting a 'unet' model (i.e. semantic segmentation),
    hence the model must be shaped as follows:
    - an input layer of shape (None, image_size, image_size, 3)
    - an output layer of shape (None, image_size, image_size, nb_labels)
    - 69 hidden layers, amonst which:
      + 57 are related to convolution
      + 8 are related to upsampling
      + 4 are related to pooling
    """
    model = geoinference.get_trained_model(
        "./tests/data/tanzania/output/semseg/checkpoints/",
        tanzania_image_size,
        tanzania_nb_labels,
    )
    assert model.input_shape[1:] == (
        tanzania_image_size,
        tanzania_image_size,
        3,
    )
    assert model.output_shape[1:] == (
        tanzania_image_size,
        tanzania_image_size,
        tanzania_nb_labels,
    )
    model_input_layer = [ml for ml in model.layers if "input" in ml.name]
    assert len(model_input_layer) == 1
    model_output_layer = [ml for ml in model.layers if "output" in ml.name]
    assert len(model_output_layer) == 1
    model_conv_layers = [ml for ml in model.layers if "conv" in ml.name]
    assert len(model_conv_layers) == 57
    model_up_layers = [ml for ml in model.layers if "up" in ml.name]
    assert len(model_up_layers) == 8
    model_pool_layers = [ml for ml in model.layers if "pool" in ml.name]
    assert len(model_pool_layers) == 4


def test_assign_label_colors():
    """Test the label colorization function, that allows to replace label IDs
    with pixel triplets
    """
    labels = [
        {"name": "foo", "color": [0, 0, 0]},
        {"name": "bar", "color": [200, 200, 200]},
    ]
    y = np.array([[[1, 1, 0], [1, 1, 0], [0, 1, 1]], [[0, 1, 1], [0, 0, 0], [0, 1, 1]]])
    expected_labels = np.array(
        [
            [
                [[200, 200, 200], [200, 200, 200], [0, 0, 0]],
                [[200, 200, 200], [200, 200, 200], [0, 0, 0]],
                [[0, 0, 0], [200, 200, 200], [200, 200, 200]],
            ],
            [
                [[0, 0, 0], [200, 200, 200], [200, 200, 200]],
                [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                [[0, 0, 0], [200, 200, 200], [200, 200, 200]],
            ],
        ]
    )
    labels = geoinference.assign_label_colors(y, labels)
    assert np.all(labels == expected_labels)


def test_extract_coordinates_from_filenames():
    """Test the x, y coordinates extraction from filenames

    The filename chosen convention is as follows:
        <path>/<name>_<width>_<height>_<x>_<y>.<extension>
    """
    filenames = [
        "./foofolder/foo_100_100_0_1000.png",
        "./barfolder/bar_100_100_2000_2100.png",
    ]
    coordinates = geoinference.extract_coordinates_from_filenames(filenames)
    assert np.all(coordinates == [[0, 1000], [2000, 2100]])


def test_fill_labelled_image():
    """Test the numpy full labelled image building

    Work with pixel tables, hence fixtures must be (n, n, 3)-shaped structures
    """
    a1 = np.array([[1, 2], [3, 4]], dtype=np.int8)
    a2 = np.array([[5, 6], [7, 8]], dtype=np.int8)
    a3 = np.array([[9, 10], [11, 12]], dtype=np.int8)
    a4 = np.array([[13, 14], [15, 16]], dtype=np.int8)
    tiles = np.array([a1, a2, a3, a4])
    x1, y1 = 0, 0
    x2, y2 = 2, 0
    x3, y3 = 0, 2
    x4, y4 = 2, 2
    coordinates = [[x1, y1], [x2, y2], [x3, y3], [x4, y4]]
    e1 = np.array(
        [[1, 2, 5, 6], [3, 4, 7, 8], [9, 10, 13, 14], [11, 12, 15, 16]],
        dtype=np.int8,
    )
    e2 = geoinference.fill_labelled_image(tiles, coordinates, 2, 4)
    assert np.all(e1 == e2)


def test_fill_labelled_image_incompatible_sizes():
    """Test the numpy full labelled image building

    Work with pixel tables, hence fixtures must be (n, n, 3)-shaped structures

    This test case corresponds to the situation where raw image dimensions are
    not multiples of the tile size.
    """
    a1 = np.array([[1, 2], [3, 4]], dtype=np.int8)
    a2 = np.array([[5, 6], [7, 8]], dtype=np.int8)
    a3 = np.array([[9, 10], [11, 12]], dtype=np.int8)
    a4 = np.array([[13, 14], [15, 16]], dtype=np.int8)
    a5 = np.array([[17, 18], [19, 20]], dtype=np.int8)
    a6 = np.array([[21, 22], [23, 24]], dtype=np.int8)
    tiles = np.array([a1, a2, a3, a4, a5, a6])
    x1, y1 = 0, 0
    x2, y2 = 2, 0
    x3, y3 = 0, 2
    x4, y4 = 2, 2
    x5, y5 = 0, 4
    x6, y6 = 2, 4
    coordinates = [[x1, y1], [x2, y2], [x3, y3], [x4, y4], [x5, y5], [x6, y6]]
    e1 = np.array(
        [[1, 2, 5], [3, 4, 7], [9, 10, 13], [11, 12, 15], [17, 18, 21]],
        dtype=np.int8,
    )
    e2 = geoinference.fill_labelled_image(tiles, coordinates, 2, 3, 5)
    assert np.all(e1 == e2)


def test_fill_labelled_image_different_sizes():
    """Test the numpy full labelled image building

    Work with label tables, hence fixtures must be (n, n)-shaped structures

    This test case corresponds to the situation where raw image width and
    height are not equal.
    """
    a1 = np.array([[1, 2], [3, 4]], dtype=np.int8)
    a2 = np.array([[5, 6], [7, 8]], dtype=np.int8)
    a3 = np.array([[9, 10], [11, 12]], dtype=np.int8)
    a4 = np.array([[13, 14], [15, 16]], dtype=np.int8)
    a5 = np.array([[17, 18], [19, 20]], dtype=np.int8)
    a6 = np.array([[21, 22], [23, 24]], dtype=np.int8)
    tiles = np.array([a1, a2, a3, a4, a5, a6])
    x1, y1 = 0, 0
    x2, y2 = 2, 0
    x3, y3 = 0, 2
    x4, y4 = 2, 2
    x5, y5 = 0, 4
    x6, y6 = 2, 4
    coordinates = [[x1, y1], [x2, y2], [x3, y3], [x4, y4], [x5, y5], [x6, y6]]
    e1 = np.array(
        [
            [1, 2, 5, 6],
            [3, 4, 7, 8],
            [9, 10, 13, 14],
            [11, 12, 15, 16],
            [17, 18, 21, 22],
            [19, 20, 23, 24],
        ],
        dtype=np.int8,
    )
    e2 = geoinference.fill_labelled_image(tiles, coordinates, 2, 4, 6)
    assert np.all(e1 == e2)


def test_build_full_labelled_image(
    tanzania_image_size, tanzania_nb_labels, tanzania_raw_image_size
):
    """Test the label prediction on a high-resolution image that has to be
        tiled during inference process

    The labelled output is composed of label IDs that must corresponds to the
    dataset glossary, and its shape must equal the original image size.
    """
    datapath = "./tests/data"
    dataset = "tanzania"
    image_paths = geoinference.get_image_paths(
        os.path.join(
            datapath, dataset, "preprocessed", str(tanzania_image_size), "testing"
        ),
        "tanzania_sample",
    )
    images = geoinference.extract_images(image_paths)
    coordinates = geoinference.extract_coordinates_from_filenames(image_paths)
    model_filename = f"best-model-{tanzania_image_size}.h5"
    model = geoinference.get_trained_model(
        os.path.join(datapath, dataset, "output/semseg/checkpoints/", model_filename),
        tanzania_image_size,
        tanzania_nb_labels,
    )
    labelled_image = geoinference.build_full_labelled_image(
        images,
        coordinates,
        model,
        tile_size=tanzania_image_size,
        img_width=tanzania_raw_image_size,
        batch_size=2,
    )
    assert labelled_image.shape == (
        tanzania_raw_image_size,
        tanzania_raw_image_size,
    )
    assert np.all([ul in range(tanzania_nb_labels) for ul in np.unique(labelled_image)])


def test_draw_grid(tanzania_image_size, tanzania_raw_image_size):
    """Test the grid drawing on a labelled image

    This function adds white (i.e. (255, 255, 255)) pixels each 'image_size'
    pixel vertically and horizontally, without modifying the input data shape.
    """
    data = np.zeros(
        [tanzania_raw_image_size, tanzania_raw_image_size, 3], dtype=np.uint8
    )
    gridded_data = geoinference.draw_grid(
        data,
        tanzania_raw_image_size,
        tanzania_raw_image_size,
        tanzania_image_size,
    )
    assert data.shape == gridded_data.shape
    expected_colors = np.array([[0, 0, 0], [255, 255, 255]], dtype=np.uint8)
    colors = np.unique(np.reshape(gridded_data, [-1, 3]), axis=0)
    assert np.all(colors == expected_colors)
    expected_white_pixels = int(tanzania_raw_image_size / tanzania_image_size)
    gridded_line = gridded_data[0]
    white_horiz_pixels = np.all(gridded_line == [255, 255, 255], axis=1)
    assert sum(white_horiz_pixels) == expected_white_pixels
    gridded_column = gridded_data[:, 0]
    white_vertic_pixels = np.all(gridded_column == [255, 255, 255], axis=1)
    assert sum(white_vertic_pixels) == expected_white_pixels


def test_get_image_features():
    """Test the image geographic feature recovering:
    - 'south', 'north', 'west' and 'east' are the image geographic coordinates,
    hence floating numbers
    - west is smaller than east
    - south is smaller than north
    - srid is an integer geocode
    - width and height are strictly positive int, as they represent the image
    size, in pixels
    - geotransform is a 6-item structure
    - wkt_projection is a string (hard to test the content itself...)
    """
    geofeatures = geoinference.get_image_features(
        "./tests/data", "tanzania", "tanzania_sample"
    )
    assert isinstance(geofeatures["south"], float)
    assert isinstance(geofeatures["north"], float)
    assert isinstance(geofeatures["east"], float)
    assert isinstance(geofeatures["west"], float)
    assert geofeatures["west"] < geofeatures["east"]
    assert geofeatures["south"] < geofeatures["north"]
    assert isinstance(geofeatures["srid"], int)
    assert isinstance(geofeatures["width"], int)
    assert isinstance(geofeatures["height"], int)
    assert geofeatures["width"] > 0
    assert geofeatures["height"] > 0
    assert len(geofeatures["geotransform"]) == 6
    assert isinstance(geofeatures["wkt_projection"], str)


def test_store_dataset():
    """Test the GDAL output dataset storage on the file system. The output raster must have the same
    properties than the original dataset used as template (whatever the output data).

    """
    input_filename = "tests/data/tanzania/input/training/images/tanzania_sample.tif"
    input_ds = gdal.Open(input_filename)
    dummydata = np.random.randint(
        0, 255, (input_ds.RasterXSize, input_ds.RasterYSize, 3)
    )
    output_label_filename = Path(
        "tests/data/tanzania/output/semseg/predicted_labels/dummylabels.tif"
    )
    features = geoinference.get_image_features(
        "tests/data", "tanzania", "tanzania_sample"
    )
    geoinference.store_raster(dummydata, output_label_filename, features)
    assert output_label_filename.exists() and output_label_filename.is_file()
    output_ds = gdal.Open(str(output_label_filename))
    assert output_ds.RasterXSize == input_ds.RasterXSize
    assert output_ds.RasterYSize == input_ds.RasterYSize
    assert output_ds.RasterCount == input_ds.RasterCount
    assert output_ds.GetGeoTransform() == input_ds.GetGeoTransform()
    assert output_ds.GetProjection() == input_ds.GetProjection()
    input_ds = None
    output_ds = None
    output_label_filename.unlink()



"""Unit test related to the dataset creation, population and loading
"""

import os
import shutil
from pathlib import Path

import pytest

from deeposlandia.datasets.aerial import AerialDataset
from deeposlandia.datasets.mapillary import MapillaryDataset
from deeposlandia.datasets.shapes import ShapeDataset
from deeposlandia.datasets.tanzania import TanzaniaDataset


def test_mapillary_dataset_creation(
    mapillary_image_size, mapillary_nb_labels, mapillary_input_config
):
    """Create a Mapillary dataset from a configuration file"""
    dataset = MapillaryDataset(mapillary_image_size, mapillary_input_config)
    assert dataset.image_size == mapillary_image_size
    assert dataset.get_nb_labels(see_all=True) == mapillary_nb_labels
    assert dataset.get_nb_images() == 0


def test_mapillary_dataset_population(
    mapillary_image_size,
    mapillary_raw_sample,
    mapillary_nb_images,
    mapillary_nb_labels,
    mapillary_input_config,
    mapillary_config,
    mapillary_temp_dir,
):
    """Populate a Mapillary dataset"""
    dataset = MapillaryDataset(mapillary_image_size, mapillary_input_config)
    dataset.populate(
        str(mapillary_temp_dir),
        mapillary_raw_sample,
        nb_images=mapillary_nb_images,
    )
    dataset.save(str(mapillary_config))
    assert dataset.get_nb_labels(see_all=True) == mapillary_nb_labels
    assert dataset.get_nb_images() == mapillary_nb_images
    assert os.path.isfile(str(mapillary_config))
    assert all(
        len(os.listdir(os.path.join(str(mapillary_temp_dir), tmp_dir)))
        == mapillary_nb_images
        for tmp_dir in ["images", "labels"]
    )


def test_mapillary_dataset_population_without_labels(
    mapillary_image_size,
    mapillary_input_config,
    mapillary_sample_without_labels_dir,
    mapillary_nb_images,
    mapillary_temp_dir,
):
    """Fail at populating a Mapillary dataset without labelled images"""
    dataset = MapillaryDataset(mapillary_image_size, mapillary_input_config)
    with pytest.raises(FileNotFoundError) as excinfo:
        dataset.populate(
            str(mapillary_temp_dir),
            mapillary_sample_without_labels_dir,
            nb_images=mapillary_nb_images,
        )
    assert (
        str(excinfo.value).split(":", maxsplit=1)[0]
        == "[Errno 2] No such file or directory"
    )


def test_mapillary_dataset_loading(
    mapillary_image_size,
    mapillary_nb_images,
    mapillary_input_config,
    mapillary_nb_labels,
    mapillary_sample_config,
):
    """Load images into a Mapillary dataset"""
    dataset = MapillaryDataset(mapillary_image_size, mapillary_input_config)
    dataset.load(mapillary_sample_config)
    assert dataset.get_nb_labels() == mapillary_nb_labels
    assert dataset.get_nb_images() == mapillary_nb_images


def test_shape_dataset_creation(shapes_image_size, shapes_nb_labels):
    """Create a Shapes dataset"""
    dataset = ShapeDataset(shapes_image_size)
    assert dataset.image_size == shapes_image_size
    assert dataset.get_nb_labels() == shapes_nb_labels
    assert dataset.get_nb_images() == 0


def test_shape_dataset_population(
    shapes_image_size,
    shapes_nb_images,
    shapes_nb_labels,
    shapes_config,
    shapes_temp_dir,
):
    """Populate a Shapes dataset"""
    dataset = ShapeDataset(shapes_image_size)
    dataset.populate(str(shapes_temp_dir), nb_images=shapes_nb_images)
    dataset.save(str(shapes_config))
    assert dataset.get_nb_labels() == shapes_nb_labels
    assert dataset.get_nb_images() == shapes_nb_images
    assert os.path.isfile(str(shapes_config))
    assert all(
        len(os.listdir(os.path.join(str(shapes_temp_dir), tmp_dir))) == shapes_nb_images
        for tmp_dir in ["images", "labels"]
    )


def test_shape_dataset_loading(
    shapes_image_size, shapes_nb_images, shapes_nb_labels, shapes_sample_config
):
    """Load images into a Shapes dataset"""
    dataset = ShapeDataset(shapes_image_size)
    dataset.load(shapes_sample_config)
    assert dataset.get_nb_labels() == shapes_nb_labels
    assert dataset.get_nb_images() == shapes_nb_images


def test_aerial_dataset_creation(aerial_image_size, aerial_nb_labels):
    """Create a AerialImage dataset"""
    dataset = AerialDataset(aerial_image_size)
    assert dataset.image_size == aerial_image_size
    assert dataset.get_nb_labels() == aerial_nb_labels
    assert dataset.get_nb_images() == 0


def test_aerial_training_dataset_population(
    aerial_image_size,
    aerial_training_temp_dir,
    aerial_raw_sample,
    aerial_nb_images,
    nb_tiles_per_image,
    aerial_training_config,
    aerial_nb_labels,
    aerial_nb_output_training_images,
):
    """Populate a Aerial dataset"""
    dataset = AerialDataset(aerial_image_size)
    dataset.populate(
        str(aerial_training_temp_dir),
        aerial_raw_sample,
        nb_images=aerial_nb_images,
        nb_tiles_per_image=nb_tiles_per_image,
    )
    dataset.save(str(aerial_training_config))
    assert dataset.get_nb_labels() == aerial_nb_labels
    assert dataset.get_nb_images() >= 0.1 * aerial_nb_output_training_images
    assert dataset.get_nb_images() <= aerial_nb_output_training_images + 3
    assert os.path.isfile(str(aerial_training_config))
    assert all(
        len(os.listdir(os.path.join(str(aerial_training_temp_dir), tmp_dir)))
        == dataset.get_nb_images()
        for tmp_dir in ["images", "labels"]
    )


def test_aerial_testing_dataset_population(
    aerial_image_size,
    aerial_testing_temp_dir,
    aerial_raw_sample,
    aerial_nb_images,
    aerial_testing_config,
    aerial_nb_labels,
    aerial_nb_output_testing_images,
):
    """Populate a Aerial dataset"""
    dataset = AerialDataset(aerial_image_size)
    dataset.populate(
        str(aerial_testing_temp_dir),
        aerial_raw_sample,
        nb_images=aerial_nb_images,
        labelling=False,
    )
    dataset.save(str(aerial_testing_config))
    assert dataset.get_nb_labels() == aerial_nb_labels
    assert dataset.get_nb_images() == aerial_nb_output_testing_images
    assert os.path.isfile(str(aerial_testing_config))
    assert (
        len(os.listdir(os.path.join(str(aerial_testing_temp_dir), "images")))
        == dataset.get_nb_images()
    )


def test_aerial_dataset_loading(
    aerial_image_size,
    aerial_testing_config,
    aerial_nb_labels,
    aerial_nb_output_testing_images,
):
    """Load images into a AerialImage dataset"""
    dataset = AerialDataset(aerial_image_size)
    dataset.load(aerial_testing_config)
    assert dataset.get_nb_labels() == aerial_nb_labels
    assert dataset.get_nb_images() == aerial_nb_output_testing_images


def test_tanzania_dataset_creation(tanzania_image_size, tanzania_nb_labels):
    """Create a Tanzania dataset"""
    dataset = TanzaniaDataset(tanzania_image_size)
    assert dataset.image_size == tanzania_image_size
    assert dataset.get_nb_labels() == tanzania_nb_labels
    assert dataset.get_nb_images() == 0


def test_tanzania_training_dataset_population(
    tanzania_image_size,
    tanzania_training_temp_dir,
    tanzania_raw_sample,
    tanzania_nb_images,
    nb_tiles_per_image,
    tanzania_training_config,
    tanzania_nb_labels,
    tanzania_nb_output_training_images,
):
    """Populate a Tanzania dataset"""
    dataset = TanzaniaDataset(tanzania_image_size)
    dataset.populate(
        str(tanzania_training_temp_dir),
        tanzania_raw_sample,
        nb_images=tanzania_nb_images,
        nb_tiles_per_image=nb_tiles_per_image,
    )
    dataset.save(str(tanzania_training_config))
    assert dataset.get_nb_labels() == tanzania_nb_labels
    assert dataset.get_nb_images() >= 0.1 * tanzania_nb_output_training_images
    assert dataset.get_nb_images() <= tanzania_nb_output_training_images + 3
    assert os.path.isfile(str(tanzania_training_config))
    assert all(
        len(os.listdir(os.path.join(str(tanzania_training_temp_dir), tmp_dir)))
        == dataset.get_nb_images()
        for tmp_dir in ["images", "labels"]
    )


def test_tanzania_testing_dataset_population(
    tanzania_image_size,
    tanzania_testing_temp_dir,
    tanzania_raw_sample,
    tanzania_nb_images,
    tanzania_testing_config,
    tanzania_nb_labels,
    tanzania_nb_output_testing_images,
    tanzania_example_image,
):
    """Populate a Tanzania dataset"""
    dataset = TanzaniaDataset(tanzania_image_size)
    dataset.populate(
        str(tanzania_testing_temp_dir),
        tanzania_raw_sample,
        nb_images=tanzania_nb_images,
        labelling=False,
    )
    dataset.save(str(tanzania_testing_config))
    assert dataset.get_nb_labels() == tanzania_nb_labels
    assert dataset.get_nb_images() == tanzania_nb_output_testing_images
    assert os.path.isfile(str(tanzania_testing_config))
    assert (
        len(os.listdir(os.path.join(str(tanzania_testing_temp_dir), "images")))
        == dataset.get_nb_images()
    )
    # Test if one may re-populate with other images: should append the new images to the dataset
    image_basename = Path(tanzania_example_image).stem
    copied_image_basename = Path(tanzania_example_image).stem.replace(
        image_basename, image_basename + "-copy"
    )
    tanzania_example_copied_image = (
        Path(tanzania_example_image).parent / copied_image_basename
    )
    shutil.copy(tanzania_example_image, tanzania_example_copied_image)
    dataset.populate(
        str(tanzania_testing_temp_dir),
        tanzania_raw_sample,
        image_basenames=[copied_image_basename],
        labelling=False,
    )
    assert dataset.get_nb_images() == 2 * tanzania_nb_output_testing_images
    assert (
        len(os.listdir(os.path.join(str(tanzania_testing_temp_dir), "images")))
        == dataset.get_nb_images()
    )
    tanzania_example_copied_image.unlink()


def test_tanzania_testing_dataset_loading(
    tanzania_image_size,
    tanzania_testing_config,
    tanzania_nb_labels,
    tanzania_nb_output_testing_images,
):
    """Load images into a Tanzania dataset"""
    dataset = TanzaniaDataset(tanzania_image_size)
    dataset.load(tanzania_testing_config)
    assert dataset.get_nb_labels() == tanzania_nb_labels
    assert dataset.get_nb_images() == tanzania_nb_output_testing_images
